import json
import logging
import os
import shutil
import tempfile

import click
import requests
import yaml


def validate_manifest(manifest):
    if 'name' not in manifest:
        raise ValueError('No name property found in manifest')

    return manifest


def read_dotfile():
    path = os.path.expanduser('~/.galileo-cli')
    cfg = None

    if os.path.isfile(path):
        with open(path, 'r') as fd:
            cfg = json.load(fd)

    if cfg:
        if 'api' in cfg:
            os.environ['galileo_api_url'] = cfg['api']


class ApiClient:

    def __init__(self, api=None) -> None:
        super().__init__()
        self.api = api or os.getenv('galileo_api_url', 'http://localhost:5001/api')

    def apps_post(self, zip_path):
        with open(zip_path, 'rb') as fd:
            response = requests.post(self._url('/apps'), data=fd.read())
            return response.json()

    def apps_delete(self, app_name):
        return requests.delete(self._url(f'/apps/{app_name}')).json()

    def apps_list(self):
        return requests.get(self._url('/apps')).json()

    def _url(self, param):
        return self.api + param


@click.group()
@click.option('--debug/--no-debug', default=False)
def galileo(debug):
    if debug:
        click.echo('Debug mode is %s' % ('on' if debug else 'off'))
        logging.basicConfig(level=logging.DEBUG)


def zipdir(path, zfd):
    # zfd is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            zfd.write(os.path.join(root, file))


@galileo.group()
def app():
    pass


@app.command('ls', help='List all available galileo apps in the repository')
def app_ls():
    apps = client.apps_list()
    for app in apps:
        click.echo(app)


@app.command('deploy', help='Package and deploy a galileo app to the repository')
@click.option('--path', required=True, help='The path to the deployable app')
def app_deploy(path):
    if path.endswith('.zip'):
        app_info = client.apps_post(path)
        click.echo(app_info)
        return 0

    if 'manifest.yml' not in os.listdir(path):
        click.echo(f'No manifest.yml found in {path}')
        return 1

    with open(os.path.join(path, 'manifest.yml'), 'r') as fd:
        y = yaml.safe_load(fd.read())

    manifest = validate_manifest(y)
    click.echo('Manifest OK: %s' % manifest['name'])

    with tempfile.TemporaryDirectory() as tmpdir:
        zip_name = os.path.join(tmpdir, 'app')
        zip_path = os.path.join(tmpdir, 'app.zip')

        # TODO: proper packaging (only what is in the manifest)
        shutil.make_archive(zip_name, 'zip', path)

        app_info = client.apps_post(zip_path)

    click.echo(app_info)


@app.command('rm', help='Remove a galileo app')
@click.argument('app-name', type=str, required=True)
def app_rm(app_name):
    response = client.apps_delete(app_name)
    if response:
        click.echo(f'deleted {app_name}')
    else:
        click.echo(f'did not delete {app_name}')


client: ApiClient


def main(*args, **kwargs):
    read_dotfile()

    global client
    client = ApiClient()
    galileo(*args, **kwargs)


if __name__ == '__main__':
    main()



import abc
import math
import random
import threading
from functools import reduce

from galileo.routing.table import RoutingTable


class Balancer(abc.ABC):

    def next_host(self, service=None):
        raise NotImplementedError


class StaticHostBalancer(Balancer):
    host: str

    def __init__(self, host) -> None:
        super().__init__()
        self.host = host

    def next_host(self, service=None):
        return self.host


class StaticLocalhostBalancer(StaticHostBalancer):

    def __init__(self) -> None:
        super().__init__('localhost')


class WeightedRandomBalancer(Balancer):
    _rtbl: RoutingTable

    def __init__(self, rtbl: RoutingTable) -> None:
        super().__init__()
        self._rtbl = rtbl

    def next_host(self, service=None):
        if not service:
            raise ValueError

        record = self._rtbl.get_routing(service)
        host = random.choices(record.hosts, record.weights, k=1)[0]

        return host


def gcd(ls):
    return reduce(math.gcd, ls)


class WeightedRoundRobinBalancer(Balancer):
    """
    Implementation of http://kb.linuxvirtualserver.org/wiki/Weighted_Round-Robin_Scheduling

    FIXME: generators are never freed
    """

    def __init__(self, rtbl: RoutingTable) -> None:
        super().__init__()
        self._rtbl = rtbl
        self._generators = dict()  # service name -> generator
        self._lock = threading.Lock()

    def generator(self, service):
        i = -1
        cw = 0

        while True:
            record = self._rtbl.get_routing(service)

            hosts = record.hosts
            weights = [int(w) for w in record.weights]

            n = len(hosts)
            i = (i + 1) % n
            if i == 0:
                cw = cw - gcd(weights)
                if cw <= 0:
                    cw = max(weights)
                    if cw == 0:
                        raise ValueError

            if weights[i] >= cw:
                yield hosts[i]

    def _require_generator(self, service):
        gen = self._generators.get(service)
        if gen:
            return gen

        with self._lock:
            if service in self._generators:  # avoid race condition
                return self._generators[service]

            gen = self.generator(service)
            self._generators[service] = gen
            return gen

    def next_host(self, service=None):
        if not service:
            raise ValueError

        gen = self._require_generator(service)
        return next(gen)



import pymq
from galileodb import Experiment

from galileo.controller import ExperimentController
from tests.api import ResourceTest


class TestExperimentsResource(ResourceTest):

    def test_get_experiments_empty(self):
        result = self.simulate_get('/api/experiments')
        self.assertEqual(result.json, [])

    def test_get_experiments(self):
        self.maxDiff = None
        experiments = [
            Experiment('id1', 'name1', 'creator1', 123, 456, 789, 'FINISH'),
            Experiment('id2', 'name2', 'creator2', 1, 4, 7, 'QUEUED'),
            Experiment('id3', 'name3', 'creator3', 3, 6, 9, 'PROGRESS')
        ]

        self.db_resource.db.save_experiment(experiments[0])
        self.db_resource.db.save_experiment(experiments[1])
        self.db_resource.db.save_experiment(experiments[2])

        result = self.simulate_get('/api/experiments')
        dicts = []
        for exp in experiments:
            dicts.append(exp.__dict__)

        self.assertCountEqual(result.json, dicts)

    def test_post_experiment(self):
        payload = {
            'experiment': {
                'name': 'my_experiment',
                'creator': 'research star'
            },
            'configuration': {
                'duration': '10s',
                'interval': '2s',
                'workloads': [
                    {
                        'service': 'alexnet',
                        'ticks': [1, 2, 3, 1, 1],
                        'arrival_pattern': 'constant',
                        'clients_per_host': 2
                    },
                ]
            }
        }

        self.ctx.cctrl.register_worker('host1')

        result = self.simulate_post('/api/experiments', json=payload)
        self.assertIsNotNone(result.json, 'Response must not be none')

        all_exps = self.db_resource.db.find_all()
        self.assertEqual(len(all_exps), 1)
        exp = all_exps[0]
        self.assertEqual(exp.id, result.json)
        self.assertEqual(exp.status, 'QUEUED')
        self.assertIsNotNone(exp.created)

        load = pymq.queue(ExperimentController.queue_key).get(timeout=2)

        # TODO test workload translation?
        self.assertIsNotNone(load)
        self.assertEqual(exp.id, load.experiment.id)

    # TODO test with no hosts available



import abc
import logging
import threading
from typing import NamedTuple, List

import redis

logger = logging.getLogger(__name__)


class RoutingRecord(NamedTuple):
    service: str
    hosts: List[str]
    weights: List[float]


class RoutingTable(abc.ABC):
    def get_routing(self, service) -> RoutingRecord:
        raise NotImplementedError

    def set_routing(self, record: RoutingRecord):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def list_services(self):
        raise NotImplementedError

    def remove_service(self, service):
        raise NotImplementedError

    def get_routes(self):
        return [self.get_routing(service) for service in self.list_services()]


class RedisRoutingTable(RoutingTable):
    update_channel = 'routing:updates'

    def __init__(self, rds) -> None:
        super().__init__()
        self.rds = rds

    def list_services(self):
        return self.rds.smembers('routing:services')

    def get_routing(self, service) -> RoutingRecord:
        rds = self.rds.pipeline()
        rds.lrange('routing:hosts:%s' % service, 0, -1)
        rds.lrange('routing:weights:%s' % service, 0, -1)
        result = rds.execute()

        if result[0]:
            return RoutingRecord(service, result[0], [float(i) for i in result[1]])

        raise ValueError(f"No routing record found for service '{service}'")

    def set_routing(self, record: RoutingRecord):
        if len(record.weights) != len(record.hosts):
            raise ValueError('The number of weights does not match the population')

        rds = self.rds.pipeline()
        rds.delete('routing:hosts:%s' % record.service)
        rds.delete('routing:weights:%s' % record.service)

        rds.sadd('routing:services', record.service)
        rds.rpush('routing:hosts:%s' % record.service, *record.hosts)
        rds.rpush('routing:weights:%s' % record.service, *record.weights)
        rds.publish(self.update_channel, record.service)

        rds.execute()

    def remove_service(self, service):
        self.remove(service)

    def clear(self):
        rds = self.rds.pipeline()
        for service in self.list_services():
            rds.delete('routing:hosts:%s' % service)
            rds.delete('routing:weights:%s' % service)
            rds.publish(self.update_channel, service)
        rds.delete('routing:services')
        rds.execute()

    def remove(self, service):
        rds = self.rds.pipeline()
        rds.delete('routing:hosts:%s' % service)
        rds.delete('routing:weights:%s' % service)
        rds.srem('routing:services', service)
        rds.publish(self.update_channel, service)
        rds.execute()


class ReadOnlyListeningRedisRoutingTable(RoutingTable):

    def __init__(self, rds) -> None:
        super().__init__()
        self.rds = rds
        self.rtable = RedisRoutingTable(rds)
        self._cache = dict()
        self._services = list()

        self._reload_lock = threading.Lock()

        self._pubsub = None
        self._thread = threading.Thread(target=self.listen)

    def start(self):
        self._thread.start()

    def stop(self, timeout=None):
        self.close()
        self._thread.join(timeout)

    def listen(self):
        self._pubsub = self.rds.pubsub()

        try:
            self._pubsub.subscribe(self.rtable.update_channel)

            with self._reload_lock:
                self._services = self.rtable.list_services()

            for item in self._pubsub.listen():
                if item['type'] == 'unsubscribe':
                    break
                if item['type'] != 'message':
                    continue

                service = item['data']

                logger.debug('received routing table update for %s', service)

                with self._reload_lock:
                    if service in self._cache:
                        del self._cache[service]
                    self._services = self.rtable.list_services()
        except redis.ConnectionError:
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception('listener terminated due to connection error')
            else:
                logger.error('listener terminated due to connection error')
        finally:
            self._pubsub.close()

    def close(self):
        try:
            self._pubsub.unsubscribe()
        except redis.ConnectionError:
            # fail silently if connection is already closed
            pass

    def list_services(self):
        return self._services

    def get_routing(self, service) -> RoutingRecord:
        record = self._cache.get(service, None)

        if record is not None:
            return record

        with self._reload_lock:
            record = self.rtable.get_routing(service)
            self._cache[service] = record
            logger.debug('loaded routing record into cache %s', record)
            return record

    def set_routing(self, record: RoutingRecord):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError

    def remove_service(self, service):
        raise NotImplementedError



import unittest
from unittest.mock import patch

from galileo.routing.balancer import StaticLocalhostBalancer
from galileo.routing.router import StaticRouter, ServiceRequest, HostRouter, ServiceRouter


class TestRouterUrlCreation(unittest.TestCase):
    def test_static_router(self):
        router = StaticRouter('http://localhost')
        url = router._get_url(ServiceRequest('foobar', '/some/service'))
        self.assertEqual('http://localhost/some/service', url)

    def test_symmetry_host_router(self):
        balancer = StaticLocalhostBalancer()
        router = HostRouter(balancer)

        url = router._get_url(ServiceRequest('foobar', '/some/service'))
        self.assertEqual('http://localhost/some/service', url)

    def test_symmetry_service_router(self):
        balancer = StaticLocalhostBalancer()
        router = ServiceRouter(balancer)

        url = router._get_url(ServiceRequest('foobar', '/some/service'))
        self.assertEqual('http://localhost/foobar/some/service', url)

    @patch('galileo.routing.router.requests.request')
    def test_request_basic(self, mock_request):
        """
        Whitebox test to check whether requests is called correctly by the router
        """

        # mock http server
        class MockResponse:
            method = 'get'
            status_code = 200
            args: dict
            kwargs: dict

        def mocked_request(*args, **kwargs):
            response = MockResponse()
            response.args = args
            response.kwargs = kwargs
            return response

        mock_request.side_effect = mocked_request

        router = StaticRouter('http://localhost')
        response = router.request(ServiceRequest('foobar', '/some/service'))

        self.assertEqual('get', response.args[0])
        self.assertEqual('http://localhost/some/service', response.args[1])


if __name__ == '__main__':
    unittest.main()



import logging
import os
import shutil
import tempfile
import zipfile
from typing import List, Optional

import requests
import yaml

from galileo.apps import AppInfo

logger = logging.getLogger(__name__)


class RepositoryException(Exception):
    pass


class InvalidArchiveError(RepositoryException):
    pass


class InvalidManifestError(RepositoryException):
    pass


def validate_manifest(manifest):
    if 'name' not in manifest:
        raise InvalidManifestError('No name property found in manifest')

    return manifest


class Repository:
    repo_path: str

    def __init__(self, path) -> None:
        super().__init__()
        self.repo_path = path

    def get_app(self, app_name) -> Optional[AppInfo]:
        for app in self.list_apps():
            if app.name == app_name:
                return app

    def add(self, zip_path) -> AppInfo:
        if not os.path.exists(self.repo_path):
            logger.debug('repo path does not exist, creating directory')
            os.mkdir(self.repo_path)

        manifest = self._get_manifest(zip_path)

        # TODO: overwrite behavior?
        archive_path = os.path.join(self.repo_path, '%s.zip' % manifest['name'])
        shutil.copy(zip_path, archive_path)

        return AppInfo(manifest['name'], manifest, archive_path)

    def delete_app(self, app_name) -> bool:
        app = self.get_app(app_name)
        if not app:
            return False

        os.remove(app.archive_path)
        return True

    def list_apps(self) -> List[AppInfo]:
        archives = self.list_archives()

        modules = list()
        for archive in archives:
            try:
                manifest = self._get_manifest(archive)
            except RepositoryException as e:
                logger.debug('Error while reading archive %s: %s', archive, e)
                continue

            info = AppInfo(manifest['name'], manifest, archive)
            modules.append(info)

        return modules

    def list_archives(self):
        return [os.path.join(self.repo_path, f) for f in os.listdir(self.repo_path) if f.endswith('.zip')]

    def _open_archive(self, archive):
        return zipfile.ZipFile(archive, 'r')

    def _get_manifest(self, archive):
        try:
            with self._open_archive(archive) as zf:
                try:
                    zf.getinfo('manifest.yml')
                except KeyError:
                    raise InvalidArchiveError('No manifest.yml found in archive %s' % archive)

                read = zf.read('manifest.yml')
                y = yaml.safe_load(read)

                return validate_manifest(y)
        except zipfile.BadZipFile as e:
            raise InvalidArchiveError('Error extracting archive %s' % archive, e)


class RepositoryClient:
    repo_host = 'http://localhost:5001'
    repo_endpoint = '/api/apps'

    def __init__(self, repo_host, repo_endpoint=None) -> None:
        super().__init__()
        self.repo_host = repo_host

        if repo_endpoint:
            self.repo_endpoint = repo_endpoint

    def exists(self, app_name) -> bool:
        resp = requests.get(self.app_url(app_name))
        if resp.status_code == 404:
            return False

        logger.debug(resp.json())
        if resp.json()['name'] == app_name:
            return True

        return False

    @property
    def repo_url(self):
        return self.repo_host + self.repo_endpoint

    def app_url(self, app_name):
        return self.repo_url + '/' + app_name

    def list(self) -> List[AppInfo]:
        response = requests.get(self.repo_url)

        apps = response.json()

        return [AppInfo(app['name'], app['manifest']) for app in apps]

    def download_app(self, app_name, directory):
        resp = requests.get(self.app_url(app_name) + '/download')

        logger.debug('processing download response for %s %s', app_name, resp)
        resp.raise_for_status()

        with tempfile.TemporaryFile() as tmp:
            tmp.write(resp.content)
            tmp.seek(0)

            with zipfile.ZipFile(tmp) as zfd:
                logger.debug('extracting %s to %s', app_name, directory)
                zfd.extractall(directory + '/' + app_name)



import logging
import multiprocessing
import threading
from typing import Dict, List

import pymq
from galileodb.trace import START, PAUSE

import galileo.worker.client as client
from galileo.controller.cluster import RedisClusterController
from galileo.worker.api import RegisterWorkerEvent, UnregisterWorkerEvent, RegisterWorkerCommand, \
    StartTracingCommand, PauseTracingCommand, CreateClientCommand, CloseClientCommand, ClientDescription, ClientConfig
from galileo.worker.context import Context

logger = logging.getLogger(__name__)


class ClientProcess(multiprocessing.Process):
    client_description: ClientDescription

    def __init__(self, ctx, trace_queue, client_description: ClientDescription) -> None:
        super().__init__(name='process-%s' % client_description.client_id,
                         target=client.run, args=(ctx, trace_queue, client_description))
        self.client_description = client_description
        self.daemon = True


class WorkerDaemon:
    """
    The worker daemon manages multiple ClientGroup processes on a host machine and exposes several interaction points
    via pymq.
    """
    name: str

    def __init__(self, ctx: Context = None, eventbus=None, ctrl=None) -> None:
        super().__init__()
        self.ctx = ctx or Context()
        self.rds = self.ctx.create_redis()
        self.eventbus = eventbus or pymq
        self.name = self.ctx.worker_name
        self.ctrl = ctrl or RedisClusterController(self.rds, self.eventbus)

        self.trace_queue = self._create_trace_queue()
        self._trace_logger = self.ctx.create_trace_logger(self.trace_queue)

        self._lock = threading.RLock()
        self._clients: Dict[str, ClientProcess] = dict()
        self._closed = threading.Event()

        self._client_id_counter = 0

    def _create_trace_queue(self):
        return multiprocessing.Queue()

    def run(self):
        self.eventbus.subscribe(self._on_create_client_command)
        self.eventbus.subscribe(self._on_close_client_command)
        self.eventbus.subscribe(self._on_register_command)
        self.eventbus.subscribe(self._on_start_tracing)
        self.eventbus.subscribe(self._on_pause_tracing)
        self.eventbus.expose(self.worker_name, 'WorkerDaemon.ping')
        self.eventbus.expose(self.create_client, 'WorkerDaemon.create_client:%s' % self.name)

        logger.debug('WorkerDaemon %s running...', self.name)
        with self._lock:
            self._trace_logger.start()
            self._register_worker()

        try:
            self._closed.wait()
        finally:
            self._unregister_worker()
            logger.debug('WorkerDaemon %s exitting', self.name)

    def worker_name(self):
        return self.name

    def create_client(self, cmd: CreateClientCommand) -> List[ClientDescription]:
        if cmd.host != self.name:
            logger.debug('ignoring CreateClientCommand sent to %s', cmd.host)
            return []

        # TODO: validate config

        logger.info('creating client %s', cmd)

        if not cmd.num:
            return []

        result = []
        for i in range(cmd.num):
            client_id = self._create_client_id(cmd.config)
            description = ClientDescription(client_id, worker=self.name, config=cmd.config)
            logger.info('creating client %s', description)
            result.append(description)

        for d in result:
            self._clients[d.client_id] = self._start_client_process(d)

        for d in result:
            self.ctrl.register_client(d)

        return result

    def _start_client_process(self, description: ClientDescription) -> ClientProcess:
        cid = description.client_id

        if cid in self._clients:
            raise ValueError('process for client %s already registered' % cid)

        process = ClientProcess(self.ctx, self.trace_queue, description)
        logger.info('starting client process %s', process)
        process.start()
        return process

    def close(self):
        with self._lock:
            logger.debug('closing clients')
            self.close_clients()

            logger.debug("closing trace logger")
            self._trace_logger.close()
            self._trace_logger.join(timeout=3)
            self._trace_logger.terminate()
            self._trace_logger.join(timeout=2)

            logger.debug("triggering exit of control loop")
            self._closed.set()

    def close_clients(self):
        for client_id in list(self._clients.keys()):
            self.close_client(client_id)

    def close_client(self, client_id):
        if client_id not in self._clients:
            return

        # TODO: need to lock around _clients
        logger.info('closing client %s', client_id)
        process = self._clients[client_id]
        del self._clients[client_id]

        self.ctrl.unregister_client(client_id)

        process.terminate()
        logger.debug('waiting on client process %s', process.name)
        process.join(3)
        logger.debug('terminated %s', process.name)

    def _register_worker(self):
        logger.info('registering name %s', self.name)
        items = self.ctx.items()
        self.ctrl.register_worker(self.name, items)
        self.eventbus.publish(RegisterWorkerEvent(self.name))

    def _unregister_worker(self):
        logger.info('unregistering name %s', self.name)
        self.ctrl.unregister_worker(self.name)
        self.eventbus.publish(UnregisterWorkerEvent(self.name))

    def _on_create_client_command(self, command: CreateClientCommand):
        if command.host != self.name:
            logger.debug('ignoring CreateClientCommand sent to %s', command.host)
            return

        logger.info('creating client %s', command)

        # TODO

    def _on_close_client_command(self, command: CloseClientCommand):
        client_id = command.client_id

        if client_id not in self._clients:
            return

        self.close_client(client_id)

    def _on_register_command(self, _: RegisterWorkerCommand):
        logger.info('received registration command')
        self._register_worker()

    def _on_start_tracing(self, _: StartTracingCommand):
        logger.info(f'received start tracing command, tracing is activated')
        self.trace_queue.put(START)

    def _on_pause_tracing(self, _: PauseTracingCommand):
        logger.info(f'received pause tracing command, tracing is deactivated')
        self.trace_queue.put(PAUSE)

    def _create_client_id(self, config: ClientConfig):
        cnt = self._client_id_counter
        self._client_id_counter += 1
        return f'{self.name}:{config.service}:{cnt}'



from galileo.controller.cluster import ClusterController, RedisClusterController
from galileo.controller.experiment import ExperimentController, ExperimentQueue, CancelError

name = 'controller'

__all__ = [
    'ExperimentController',
    'ExperimentQueue',
    'CancelError',
    'ClusterController',
    'RedisClusterController'
]



"""
Module to start webapp from WSGI context.


"""
import os

import falcon
import pymq
import redis
from galileodb.factory import create_experiment_database_from_env
from pymq.provider.redis import RedisConfig

from galileo.apps.repository import Repository
from galileo.controller import ExperimentController, RedisClusterController
from galileo.experiment.service import SimpleExperimentService
from galileo.webapp.app import AppContext, CORSComponent, setup


def create_context() -> AppContext:
    context = AppContext()

    context.rds = redis.Redis(
        host=os.getenv('galileo_redis_host', 'localhost'),
        port=int(os.getenv('galileo_redis_port', 6379)),
        decode_responses=True
    )
    pymq.init(RedisConfig(context.rds))

    context.ectrl = ExperimentController(context.rds)
    context.cctrl = RedisClusterController(context.rds)
    context.exp_db = create_experiment_database_from_env()
    context.exp_service = SimpleExperimentService(context.exp_db)
    context.repository = Repository(os.getenv('galileo_apps_repo_dir', os.path.abspath('./apps-repo')))

    return context


api = falcon.API(middleware=[CORSComponent()])
setup(api, create_context())



import time
import unittest

from timeout_decorator import timeout_decorator

from galileo.routing.table import RoutingTable, RoutingRecord, RedisRoutingTable, ReadOnlyListeningRedisRoutingTable
from tests.testutils import RedisResource


class TestRoutingTable(unittest.TestCase):
    redis = RedisResource()
    rtbl: RoutingTable

    def setUp(self) -> None:
        self.redis.setUp()
        self.rtbl = RedisRoutingTable(self.redis.rds)

    def tearDown(self) -> None:
        self.rtbl.clear()
        self.redis.tearDown()

    def test_get_routes_on_empty_table_returns_empty_list(self):
        routes = self.rtbl.get_routes()
        self.assertEqual(0, len(routes))

    def test_set_routing(self):
        record = RoutingRecord('aservice', ['ahost', 'bhost'], [1.0, 2.0])
        self.rtbl.set_routing(record)

        routes = self.rtbl.get_routes()
        self.assertEqual(1, len(routes))
        self.assertEqual(record, routes[0])

    def test_list_services(self):
        self.rtbl.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [1.0, 2.0]))
        self.rtbl.set_routing(RoutingRecord('bservice', ['ahost', 'bhost'], [1.0, 2.0]))

        services = self.rtbl.list_services()
        self.assertIn('aservice', services)
        self.assertIn('bservice', services)
        self.assertEqual(2, len(services))

    def test_remove_service(self):
        self.rtbl.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [1.0, 2.0]))
        self.rtbl.set_routing(RoutingRecord('bservice', ['ahost', 'bhost'], [2.0, 3.0]))

        self.rtbl.remove_service('aservice')

        routes = self.rtbl.get_routes()
        self.assertEqual(1, len(routes))
        self.assertEqual('bservice', routes[0].service)

        services = self.rtbl.list_services()
        self.assertNotIn('aservice', services)

    def test_get_routing_of_non_existing_service(self):
        self.rtbl.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [1.0, 2.0]))
        self.assertRaises(ValueError, self.rtbl.get_routing, 'bservice')

    def test_set_routing_invalid_record(self):
        record = RoutingRecord('aservice', ['ahost', 'bhost'], [1.0])
        self.assertRaises(ValueError, self.rtbl.set_routing, record)
        self.assertEqual(0, len(self.rtbl.get_routes()))

        record = RoutingRecord('aservice', ['ahost'], [1.0, 2.0])
        self.assertRaises(ValueError, self.rtbl.set_routing, record)
        self.assertEqual(0, len(self.rtbl.get_routes()))

    @timeout_decorator.timeout(5)
    def test_notification_on_set(self):
        pubsub = self.redis.rds.pubsub()

        pubsub.subscribe(RedisRoutingTable.update_channel)

        listener = pubsub.listen()

        next(listener)  # subscription message

        self.rtbl.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [2.0, 3.0]))
        msg = next(listener)  # subscription message
        self.assertEqual('aservice', msg['data'])

        self.rtbl.set_routing(RoutingRecord('bservice', ['ahost', 'bhost'], [2.0, 3.0]))
        msg = next(listener)  # subscription message
        self.assertEqual('bservice', msg['data'])

        self.rtbl.remove_service('bservice')
        msg = next(listener)  # subscription message
        self.assertEqual('bservice', msg['data'])

        pubsub.unsubscribe(RedisRoutingTable.update_channel)
        pubsub.close()


class TestListeningTable(unittest.TestCase):
    redis = RedisResource()
    rtbl_mutable: RoutingTable

    def setUp(self) -> None:
        self.redis.setUp()
        self.rtbl_mutable = RedisRoutingTable(self.redis.rds)
        self.rtbl = ReadOnlyListeningRedisRoutingTable(self.redis.rds)
        self.rtbl.start()

    def tearDown(self) -> None:
        self.rtbl_mutable.clear()
        self.rtbl.stop(2)
        self.redis.tearDown()

    def test_cache_returns_same_object(self):
        self.rtbl_mutable.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [2.0, 3.0]))

        record1 = self.rtbl.get_routing('aservice')
        record2 = self.rtbl.get_routing('aservice')

        self.assertEqual(id(record1), id(record2))

    def test_get_routing_after_update_returns_correct_value(self):
        # FIXME use polling

        self.rtbl_mutable.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [2.0, 3.0]))
        record1 = self.rtbl.get_routing('aservice')

        self.rtbl_mutable.set_routing(RoutingRecord('aservice', ['ahost'], [2.0]))
        time.sleep(0.25)
        record2 = self.rtbl.get_routing('aservice')

        self.assertNotEqual(id(record1), id(record2))
        self.assertEqual('ahost', record2.hosts[0])
        self.assertEqual(1, len(record2.hosts))

    def test_list_services_returns_correct_services(self):
        # FIXME use polling

        self.rtbl_mutable.set_routing(RoutingRecord('aservice', ['ahost', 'bhost'], [2.0, 3.0]))
        time.sleep(0.25)
        self.assertIn('aservice', self.rtbl.list_services())
        self.assertEqual(1, len(self.rtbl.list_services()))

        self.rtbl_mutable.set_routing(RoutingRecord('bservice', ['ahost', 'bhost'], [2.0, 3.0]))
        time.sleep(0.25)
        self.assertIn('aservice', self.rtbl.list_services())
        self.assertIn('bservice', self.rtbl.list_services())
        self.assertEqual(2, len(self.rtbl.list_services()))

        self.rtbl_mutable.remove_service('aservice')
        time.sleep(0.25)
        self.assertIn('bservice', self.rtbl.list_services())
        self.assertEqual(1, len(self.rtbl.list_services()))


if __name__ == '__main__':
    unittest.main()



import logging
from typing import NamedTuple

logger = logging.getLogger(__name__)


class AppRequest(NamedTuple):
    app_name: str
    method: str
    endpoint: str
    kwargs: dict


class AppClient:

    def __init__(self, name, context, module) -> None:
        super().__init__()
        self.name = name
        self.context = context
        self.module = module

    def __getattr__(self, item):
        return getattr(self.module, item)

    def next_request(self) -> AppRequest:
        method, endpoint, kwargs = self.module.next_request()
        return AppRequest(self.name, method, endpoint, kwargs or {})


class DefaultAppClient(AppClient):
    """
    The default app client is a flexible http request client. It only requires the client parameters which are used as
    follows: parameters['method'] holds the HTTP method to use (defaults to 'get'), parameters['path'] holds the path to
    which the request is sent (defaults to '/'), and parameters['kwargs'] holds a dict with keyword arguments passed to
    the python requests library when performing the `request` call (e.g., 'data': ..., 'json', ...).
    """

    def __init__(self, parameters=None) -> None:
        if parameters:
            method = parameters.get('method', 'get')
            path = parameters.get('path', '/')
            kwargs = parameters.get('kwargs')
        else:
            method = 'get'
            path = '/'
            kwargs = None

        client = HttpClient(method, path, kwargs)

        super().__init__('http', None, client)


class HttpClient:
    def __init__(self, method='get', path='/', parameters=None) -> None:
        super().__init__()
        self.method = method
        self.path = path
        self.parameters = parameters or {}

    def next_request(self):
        return self.method, self.path, self.parameters



import abc
import importlib.util
import logging
import os
from typing import List

import yaml

from galileo.apps import AppInfo, Context
from galileo.apps.app import AppClient
from galileo.apps.repository import RepositoryClient

logger = logging.getLogger(__name__)


class AppClientLoader(abc.ABC):
    def list(self) -> List[AppInfo]:
        raise NotImplementedError

    def load(self, name, parameters=None) -> AppClient:
        raise NotImplementedError


class AppClientDirectoryLoader(AppClientLoader):
    _manifest_file = 'manifest.yml'
    _module_file = 'generator.py'

    root: str

    def __init__(self, root: str) -> None:
        super().__init__()
        self.root = root

    def list(self) -> List[AppInfo]:
        root = self.root
        manifest_file = self._manifest_file

        result = list()
        for d in os.listdir(root):
            manifest_path = os.path.join(root, d, manifest_file)
            if not os.path.exists(manifest_path):
                continue

            try:
                manifest = self._load_manifest(manifest_path)
                try:
                    self._validate_manifest(manifest)
                except ValueError as e:
                    logger.debug('Error validating manifest: %s', e)
                    continue

                app_info = AppInfo(manifest['name'], manifest)
                result.append(app_info)
            except:
                logger.exception('Exception loading app manifest %s', manifest_path)

        return result

    def load(self, name, parameters=None) -> AppClient:
        manifest = self._require_manifest(name)
        spec, module = self._require_module(name)

        context = Context()
        context.path = os.path.join(self.root, name)
        context.name = name
        context.manifest = manifest
        context.parameters = parameters or dict()

        module.context = context

        spec.loader.exec_module(module)

        return AppClient(name, context, module)

    def _require_module(self, name):
        module_path = os.path.join(self.root, name, self._module_file)
        spec = importlib.util.spec_from_file_location('galileo.clients.' + name, module_path)
        module = importlib.util.module_from_spec(spec)
        return spec, module

    def _require_manifest(self, name):
        manifest_path = os.path.join(self.root, name, self._manifest_file)
        if not os.path.exists(manifest_path):
            raise ValueError('No manifest.yml found in %s' % manifest_path)

        manifest = self._load_manifest(manifest_path)
        self._validate_manifest(manifest)
        return manifest

    def _load_manifest(self, file):
        with open(file) as fd:
            return yaml.safe_load(fd)

    def _validate_manifest(self, manifest):
        if 'name' not in manifest:
            raise ValueError('Manifest error: no app name specified')


class AppRepositoryFallbackLoader(AppClientLoader):
    loader: AppClientDirectoryLoader
    repo: RepositoryClient

    def __init__(self, loader, repo) -> None:
        super().__init__()
        self.loader = loader
        self.repo = repo

    def list(self) -> List[AppInfo]:
        try:
            apps = {info.name: info for info in self.repo.list()}
        except IOError:
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception('Error getting list from app repository')
            apps = {}

        for app in self.loader.list():
            # we prioritize local apps
            apps[app.name] = app

        return list(apps.values())

    def load(self, name, parameters=None) -> AppClient:
        try:
            logger.debug('trying to load app %s from filesystem %s', name, self.loader.root)
            return self.loader.load(name, parameters)
        except ValueError:
            logger.debug('Failed to load app %s locally, trying to download...', name)

        if self.repo.exists(name):
            self.repo.download_app(name, self.loader.root)
            app = self.loader.load(name, parameters)
            logger.debug('successfully loaded app %s: %s', name, app)
            return app
        else:
            raise ValueError('No app with name %s found' % name)



import unittest
from unittest.mock import MagicMock

from galileo.shell.shell import ClientGroup
from galileo.worker.api import ClientDescription


class TestClientGroup(unittest.TestCase):

    def test_add_groups(self):
        ctrl = MagicMock()

        cg1 = ClientGroup(ctrl, [
            ClientDescription('id1', 'worker1', None),
            ClientDescription('id2', 'worker1', None),
            ClientDescription('id3', 'worker2', None),
        ])

        cg2 = ClientGroup(ctrl, [
            ClientDescription('id4', 'worker1', None),
            ClientDescription('id5', 'worker2', None),
            ClientDescription('id6', 'worker2', None),
        ])

        cg3 = cg1 + cg2

        self.assertEqual(6, len(cg3.clients))

        self.assertEqual(cg1.clients[0], cg3.clients[0])
        self.assertEqual(cg2.clients[0], cg3.clients[3])
        self.assertEqual(ctrl, cg1.ctrl)
        self.assertEqual(ctrl, cg2.ctrl)

    def test_add_groups_with_different_ctrl_raises_exception(self):
        ctrl1 = MagicMock()
        ctrl2 = MagicMock()

        cg1 = ClientGroup(ctrl1, [
            ClientDescription('id1', 'worker1', None),
            ClientDescription('id2', 'worker1', None),
            ClientDescription('id3', 'worker2', None),
        ])

        cg2 = ClientGroup(ctrl2, [
            ClientDescription('id4', 'worker1', None),
            ClientDescription('id5', 'worker2', None),
            ClientDescription('id6', 'worker2', None),
        ])

        try:
            _ = cg1 + cg2
            self.fail('expected value error because of differing controllers in client groups')
        except ValueError:
            pass



import os
import shutil
import tempfile

import redislite
from galileodb.sql.adapter import ExperimentSQLDatabase, SqlAdapter
from galileodb.sql.driver.sqlite import SqliteAdapter

from galileo.util import poll


class TestResource(object):

    def setUp(self):
        pass

    def tearDown(self):
        pass


class RedisResource(TestResource):
    tmpfile: str
    rds: redislite.Redis

    def setUp(self):
        self.tmpfile = tempfile.mktemp('.db', 'galileo_test_')
        self.rds = redislite.Redis(self.tmpfile, decode_responses=True)
        self.rds.get('dummykey')  # run a first command to initiate

    def tearDown(self):
        self.rds.shutdown()

        os.remove(self.tmpfile)
        os.remove(self.rds.redis_configuration_filename)
        os.remove(self.rds.settingregistryfile)
        shutil.rmtree(self.rds.redis_dir)

        self.rds = None
        self.tmpfile = None


class SqliteResource(TestResource):
    db_file = None

    sql: SqlAdapter
    db: ExperimentSQLDatabase

    def setUp(self) -> None:
        self.db_file = tempfile.mktemp('.sqlite', 'galileo_test_')
        self.sql = SqliteAdapter(self.db_file)
        self.db = ExperimentSQLDatabase(self.sql)
        self.db.open()

    def tearDown(self) -> None:
        self.db.close()
        os.remove(self.db_file)


def assert_poll(condition, msg='Condition failed', timeout=2):
    try:
        poll(condition, timeout, 0.01)
    except TimeoutError:
        raise AssertionError(msg)



import io


class Stringer:

    def __init__(self, obj) -> None:
        super().__init__()
        self.obj = obj

    @staticmethod
    def to_string(obj):
        if isinstance(obj, str):
            return obj

        if isinstance(obj, dict):
            return Stringer.to_string([f'{k} = {v}' for k, v in obj.items()])

        if isinstance(obj, (list, tuple)):
            p = StringPrinter()
            for item in obj:
                p.print(item)
            return str(p)

        if callable(obj):
            return Stringer.to_string(obj())

        return str(obj)

    def __str__(self):
        return Stringer.to_string(self.obj).strip()

    def __repr__(self):
        return self.__str__()


class StringPrinter:

    def __init__(self) -> None:
        super().__init__()
        self.out = io.StringIO()

    def print(self, *args, **kwargs):
        print(*args, file=self.out, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.print(*args, **kwargs)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.out.getvalue()


def print_tabular(data, columns=None, widths=None, printer=None):
    if not data and not columns:
        return

    printer = printer or print
    columns = columns or data[0].keys()
    if not widths:
        widths = list()
        for c in columns:
            max_len = len(c)

            for row in data:
                max_len = max(max_len, len(str(row[c])))

            widths.append(-max_len)

    sep = ['-' * abs(i) for i in widths]
    sep = '+-' + '-+-'.join(sep) + '-+'

    row_fmt = ['%%%ds' % widths[i] for i in range(len(widths))]
    row_fmt = '| ' + ' | '.join(row_fmt) + ' |'

    header = tuple(columns)

    printer(sep)
    printer(row_fmt % header)
    printer(sep)

    for record in data:
        row = tuple([record[k] for k in columns])
        printer(row_fmt % row)

    printer(sep)


def sprint_routing_table(table) -> str:
    printer = StringPrinter()

    records = [table.get_routing(service) for service in table.list_services()]

    w = [-25, 20, 9]  # TODO: read from records

    sep = ['-' * abs(i) for i in w]
    sep = '+-' + '-+-'.join(sep) + '-+'

    row_fmt = ['%%%ds' % w[i] for i in range(len(w))]
    row_fmt = '| ' + ' | '.join(row_fmt) + ' |'

    header = ('Service', 'Hosts', 'Weights')

    printer(sep)
    printer(row_fmt % header)
    printer(sep)

    for record in records:
        for i in range(len(record.hosts)):
            ls = (record.service if i == 0 else '', record.hosts[i], record.weights[i])
            printer(row_fmt % ls)
        printer(sep)

    return str(printer)



import atexit
import multiprocessing
import os
import sys
import time
from threading import Thread, Condition
from typing import List, Dict, NamedTuple, Set, Tuple

import pymq
from galileodb.cli.recorder import run as run_recorder
from galileodb.model import Event as ExperimentEvent
from galileodb.reporter.events import RedisEventReporter as ExperimentEventReporter
from pymq.provider.redis import RedisConfig
from telemc import TelemetryController

from galileo.controller.cluster import ClusterController, RedisClusterController
from galileo.routing import RoutingRecord, RoutingTable, RedisRoutingTable
from galileo.shell.printer import sprint_routing_table, print_tabular, Stringer
from galileo.worker.api import ClientConfig, ClientDescription, CloseClientCommand, ClientInfo, WorkloadDoneEvent
from galileo.worker.client import single_request

prompt = 'galileo> '

banner = r"""
                                   __  __
 .-.,="``"=.          ____ _____ _/ (_) /__  ____
 '=/_       \        / __ `/ __ `/ / / / _ \/ __ \
  |  '=._    |      / /_/ / /_/ / / / /  __/ /_/ /
   \     `=./`.     \__, /\__,_/_/_/_/\___/\____/
    '=.__.=' `='   /____/

"""

usage = Stringer('''the galileo shell is an interactive python shell that provides the following commands

Commands:
  usage         show this message
  env           show environment variables
  pwd           show the current working directory

Functions:
  sleep         time.sleep wrapper

Objects:
  g             Galileo object that allows you to interact with the system
  show          Prints runtime information about the system to system out
  exp           Galileo experiment
  rtbl          Symmetry routing table
  telemd        Telemd object to pause, unpause and list registered telemd daemons

Type help(<function>) or help(<object>) to learn how to use the functions.
''')


def sleep(secs: float):
    """
    Delay execution for a given number of seconds. The argument may be a floating point number for subsecond precision.
    :param secs: seconds to sleep
    :return: None
    """
    time.sleep(secs)


@Stringer
def env():
    return dict(os.environ)


@Stringer
def pwd():
    return os.path.abspath(os.curdir)


class RequestFuture:
    ctrl: ClusterController
    client_ids: Set[str]

    def __init__(self, ctrl, client_ids) -> None:
        super().__init__()
        self.ctrl = ctrl
        self.client_ids = client_ids
        self.done = False
        self.aborted = False
        self.lock = Condition()

    def run(self, n=None, ia=None):
        clients_done = set(self.client_ids)

        # lots of problems with this unfortunately, may never terminate if clients disappear, concurrent events from
        # previous (aborted) calls will interfere with future calls, etc.

        def done_subscriber(event: WorkloadDoneEvent):
            clients_done.remove(event.client_id)
            if len(clients_done) == 0:
                with self.lock:
                    self.done = True
                    self.lock.notify_all()

        try:
            with self.lock:
                pymq.subscribe(done_subscriber)
                for c in self.client_ids:
                    self.ctrl.set_workload(c, ia, n)

                self.lock.wait_for(self.stopped)
        finally:
            pymq.unsubscribe(done_subscriber)

    def stopped(self):
        return self.done or self.aborted

    def abort(self):
        with self.lock:
            self.aborted = True
            self.lock.notify_all()

    def wait(self, timeout=None, abort_after_timeout=True):
        with self.lock:
            result = self.lock.wait_for(self.stopped, timeout=timeout)

        if not result and abort_after_timeout:
            self.abort()  # make sure run terminates as well


class ClientGroup:
    """
    Allows the interaction with a set of galileo clients. Functions like `rps`, `close` operate on all clients in the
    ClientGroup.
    """

    ctrl: ClusterController
    clients: List[ClientDescription]

    def __init__(self, ctrl, clients, cfg=None) -> None:
        super().__init__()
        self.ctrl = ctrl
        self.clients: List[ClientDescription] = clients
        self.cfg = cfg

        self.running_request: RequestFuture = None

    def rps(self, n: float):
        """
        Set the request generation rate of all clients in the group.
        :param n: requests per second (can be fractional, e.g. 0.5 to send a request every 2 seconds)
        """
        if n == 0:
            self.pause()
        else:
            self.request(ia=(1 / n))

    def request(self, n=None, ia=None) -> RequestFuture:
        """
        Tell the clients in the group to start generating requests. You can specify a message rate, or a number of
        requests, or both.

        The interarrival can be specified either as int/float, then it is interpreted as constant interarrival. The
        following example causes clients to send 5 message per second::

            c.request(ia=0.2)

        It can also be specified as a tuple, where the first element specifies the distribution used from ``random``:
        ``random.<distribution>``, and the remaining elements specify the arguments that are passed to the call::

            c.requests(ia=('expovariate', 1))

        Uses ``random.expovariate(1)`` to generate interarrivals.

        The messages limit can be used with or without the ``ia`` parameter::

            c.requests(n=100)

        will send 100 messages as fast as possible (clients currently work synchronously). Whereas::

            c.requests(n=100, ia=0.2)

        Will send 100 messages but pause for 0.2 between each message.

        The method returns a RequestFuture on which you can call ``wait()`` if you want to block until the clients are
        done.

            c.request(n=100).wait()

        :param n: the maximum number of requests
        :param ia: the request interarrival
        :return a RequestFuture object
        """

        # if a previous request is running, abort the future to unsubscribe from events properly
        if self.running_request and not self.running_request.stopped():
            self.running_request.abort()
            self.running_request.wait(1)

        future = RequestFuture(self.ctrl, {c.client_id for c in self.clients})
        t = Thread(target=future.run, args=(n, ia))
        t.start()

        self.running_request = future

        return future

    def pause(self):
        """
        Pause the current workload set by ``ClientGroup.request``.
        """
        for c in self.clients:
            self.ctrl.stop_workload(c.client_id)

    def info(self) -> List[ClientInfo]:
        return pymq.stub('Client.get_info', timeout=2, multi=True)()

    def add(self, n=1):
        """
        Add n number of clients with the same configuration to this ClientGroup
        :param n: the number of clients
        :return: a client group containing only the clients that were added
        """
        if self.cfg is None:
            raise ValueError('need config to add clients to group')

        new_clients = self.ctrl.create_clients(self.cfg, n)
        self.clients.extend(new_clients)
        return ClientGroup(self.ctrl, new_clients, self.cfg)

    def close(self, n=None):
        """
        Close all or a specific number of clients in the client group. Clients are terminated remotely.

        :param n: the number of clients to terminate (all if None)
        :return: the ids of the clients that were closed
        """
        if n is None:
            n = len(self.clients)

        print('closing %d clients' % n)

        removed = list()
        for _ in range(n):
            removed.append(self.clients.pop())

        for c in removed:
            pymq.publish(CloseClientCommand(c.client_id))

        return [description.client_id for description in removed]

    def __repr__(self) -> str:
        if not self.clients:
            return 'empty client group'

        return '\n'.join([c.client_id for c in self.clients])

    def __add__(self, other) -> 'ClientGroup':
        return ClientGroup.merge(self, other)

    @staticmethod
    def merge(c1: 'ClientGroup', c2: 'ClientGroup') -> 'ClientGroup':
        if c1.ctrl is not c2.ctrl:
            raise ValueError('client groups have different controllers')

        clients = list()
        clients.extend(c1.clients)
        clients.extend(c2.clients)

        return ClientGroup(c1.ctrl, clients)


class Telemd:
    telemd_ctrl: TelemetryController = None

    def __init__(self, telemd_ctrl) -> None:
        super().__init__()
        self.telemd_ctrl = telemd_ctrl

    def start_telemd(self, hosts: List[str] = None):
        """
        Send a unpause message to all registered telemd hosts
        """
        if hosts is not None:
            for host in hosts:
                self.telemd_ctrl.unpause(host)
        else:
            self.telemd_ctrl.unpause_all()

    def stop_telemd(self, hosts: List[str] = None):
        """
        Send a pause message to all registered telemd hosts
        """
        if hosts is not None:
            for host in hosts:
                self.telemd_ctrl.pause(host)
        else:
            self.telemd_ctrl.pause_all()

    def list_telemd_hosts(self):
        """List all registered telemd hosts"""
        return self.telemd_ctrl.get_nodes()


class Galileo:
    ctrl: ClusterController = None

    def __init__(self, ctrl) -> None:
        super().__init__()
        self.ctrl = ctrl

    def start_tracing(self):
        """
        Send a StartTracing command to all workers.
        :return: the number of workers who received the command
        """
        return self.ctrl.start_tracing()

    def stop_tracing(self):
        """
        Send a StopTracing command to all workers.
        :return: the number of workers who received the command
        """
        return self.ctrl.stop_tracing()

    def workers(self) -> List[str]:
        """
        List registered workers.
        :return: a list of strings
        """
        return self.ctrl.list_workers()

    def workers_info(self) -> List[Tuple[str, str]]:
        return self.ctrl.list_workers_info()

    def ping(self):
        """
        Send a synchronous ping to all workers and return the response. May contain error tuples.
        :return:
        """
        return self.ctrl.ping()

    def discover(self):
        """
        Send a discover command to all workers, which tells them to re-register to redis.
        :return: the number of workers who received the command
        """
        return self.ctrl.discover()

    def clients(self, *client_ids) -> ClientGroup:
        """
        Returns a ClientGroup for the given client_ids, or all clients of no client_ids are specified.
        :param client_ids: optional variadic client_ids
        :return: a ClientGroup
        """
        clients = self.ctrl.list_clients()

        if client_ids:
            clients = [c for c in clients if c.client_id in client_ids]

        return ClientGroup(self.ctrl, clients)

    def request(self, service, client: str = None, parameters: dict = None, router_type='SymmetryHostRouter',
                worker_labels: dict = None):
        """
        Send a request with the given configuration like a client would. See ``spawn`` for the parameters. An additional
        parameter is the ``router_type`` which specifies which router to use (see `Context.create_router`)
        """
        cfg = ClientConfig(service, client=client, parameters=parameters, worker_labels=worker_labels)
        resp = single_request(cfg, router_type=router_type)

        return resp

    def spawn(self, service, num: int = 1, client: str = None, parameters: dict = None,
              worker_labels: dict = None) -> ClientGroup:
        """
        Spawn clients for the given service and distribute them across workers. If no client app is specified, a default
        http client will be created that creates http requests from the (optional) parameters::

            spawn('myservice', parameters={'method': 'get', 'path': '/', 'kwargs': None}

        is equivalent to::

            spawn('myservice')

        Another example::

            spawn('myservice', 2, parameters={'method': 'post', 'path': '/', 'kwargs': {'data': 'my post data'}})

        will result is POST requests to the path '/' where the 'kwargs' dict is passed to the python
        ``requests.request`` call as keyword arguments.

        :param service: the service name
        :param num: the number of clients
        :param client: the client app name (optional, if not given will use service name)
        :param parameters: parameters for the app (optional, e.g.: '{ "size": "small" }'
        :param worker_labels: labels that workers must match to be part of the group
        :return a new ClientGroup for the created clients
        """
        cfg = ClientConfig(service, client=client, parameters=parameters, worker_labels=worker_labels)
        clients = self.ctrl.create_clients(cfg, num)
        return ClientGroup(self.ctrl, clients, cfg)


class ExperimentArguments(NamedTuple):
    name: str = None
    creator: str = None


class Experiment:
    event_reporter: ExperimentEventReporter

    def __init__(self, rds) -> None:
        super().__init__()
        self.event_reporter = ExperimentEventReporter(rds)
        self.experiment = None
        self._atexit = False

    def event(self, name: str, value: str = None):
        """
        Sends a galileo experiment event.

        :param name: the event name
        :param value: the optional event value
        """
        ts = time.time()
        self.event_reporter.report(ExperimentEvent(ts, name, value))
        if value:
            print(f'%.3f: %s = %s' % (ts, name, value))
        else:
            print(f'%.3f: %s' % (ts, name))

    def start(self, name=None, creator=None):
        if self.experiment is not None:
            raise ValueError('experiment already running')

        if not self._atexit:
            self._atexit = True
            atexit.register(self.stop)

        args = ExperimentArguments(name, creator)

        # FIXME: i don't really like to re-use the cli module, but it was the easiest to get it working. plus we have
        #  the experiment daemon, experiment runner, experiment recorder, ...
        self.experiment = multiprocessing.Process(target=run_recorder, args=(args,), daemon=True)
        self.experiment.start()

        print('experiment started')

    def stop(self, wait=5):
        if self.experiment is None:
            return

        try:
            self.experiment.terminate()
            print('waiting for experiment to end...')
            self.experiment.join(wait)
            print('ok')
        finally:
            self.experiment = None


class Show:
    g: Galileo

    def __init__(self, gal: Galileo) -> None:
        super().__init__()
        self.g = gal

    def workers(self):
        workers = self.g.workers()
        if workers:
            return Stringer(workers)
        else:
            return Stringer('no available workers')

    def clients(self):
        cs = self.g.clients().clients
        data = [c._asdict() for c in cs]
        print_tabular(data)

    def info(self, **exclude):
        """
        Print tabular info about the currently running clients.
        :param exclude: kwargs that exclude fields, e.g., parameters=False to hide the parameter column
        """
        data = []
        for info in self.g.clients().info():

            param_str = str(info.description.config.parameters)
            if len(param_str) > 80:
                param_str = param_str[:75] + ' ...}'

            record = {
                'client id': info.description.client_id,
                'worker': info.description.worker,
                'service': info.description.config.service,
                'client': info.description.config.client or 'default',
                'parameters': param_str,
                'requests': info.requests,
                'failed': info.failed,
            }

            for k, v in exclude.items():
                if k in record and v is False:
                    del record[k]

            data.append(record)

        print_tabular(data)


class RoutingTableHelper:
    """
    View or update the symmetry RoutingTable to control where galileo clients send their requests.
    """
    table: RoutingTable

    def __init__(self, table) -> None:
        super().__init__()
        self.table = table

    def record(self, service):
        """
        Get the RoutingRecord for a specific service.
        :param service: the service
        :return: the routing record, or None
        """
        try:
            return self.table.get_routing(service)
        except ValueError:
            return None

    def records(self):
        """
        Return all routing records
        :return: a list of RoutingRecord tuples
        """
        return [self.table.get_routing(service) for service in self.table.list_services()]

    def set(self, service: str, hosts: List[str], weights: List[float] = None):
        """
        Creates a new or overwrites an existing RoutingRecord
        :param service: the service
        :param hosts: a list of hosts
        :param weights: an optional list of weights for the hosts
        :return:
        """
        if weights is None:
            weights = [1] * len(hosts)

        record = RoutingRecord(service, hosts, weights)
        self.table.set_routing(record)
        return record

    def update_weights(self, service, weights: List[float]):
        """
        Update the weights for the routing record of the given service. raises a ValueError if the record does not
        exist, or if the number of weights does not match the number of hosts in the record.

        :param service: the service
        :param weights: the updated weights
        :return:
        """
        record = self.table.get_routing(service)

        if len(weights) != len(record.weights):
            raise ValueError('invalid number of weights')

        record.weights.clear()
        record.weights.extend(weights)

        self.table.set_routing(record)

        return self

    def append(self, service: str, host: str, weight: float = 1):
        """
        Adds a single host and optional weight to the routing record of the services. If the record does not exist, it
        is created.
        :param service: the service
        :param host: the host
        :param weight: the weight
        :return:
        """
        try:
            record = self.table.get_routing(service)
        except ValueError:
            return self.set(service, [host], [weight])

        record.hosts.append(host)
        record.weights.append(weight)
        self.table.set_routing(record)

        return self

    def remove(self, service: str):
        """
        Removes the routing record of the given service.
        :param service: the service
        :return: this object
        """
        self.table.remove_service(service)
        return self

    def clear(self):
        """
        Removes all routing records from the table.
        :return: this object
        """
        self.table.clear()
        return self

    def print(self):
        print(self.__repr__())

    def __repr__(self):
        return sprint_routing_table(self.table)


is_interactive = sys.__stdin__.isatty()


def init(rds) -> Dict[str, object]:
    eventbus = pymq.init(RedisConfig(rds))

    g = Galileo(RedisClusterController(rds))
    show = Show(g)
    rtbl = RoutingTableHelper(RedisRoutingTable(rds))
    exp = Experiment(rds)
    telemd = Telemd(TelemetryController(rds))

    return {
        'g': g,
        'show': show,
        'rtbl': rtbl,
        'exp': exp,
        'eventbus': eventbus,
        'telemd': telemd
    }


def init_module(rds, name=None):
    gvars = init(rds)

    import sys
    if name is None:
        name = __name__

    module = sys.modules[name]

    for name, value in gvars.items():
        setattr(module, name, value)



import os
import unittest
import zipfile
from tempfile import TemporaryDirectory

from galileo.apps.repository import Repository

this_dir = os.path.join(os.path.dirname(__file__))


class TemporaryRepositoryResource:
    """
    Prepares a Repository, by zipping all test apps into a temporary directory, and using that directory as root for the
    Repository.
    """
    apps_dir = os.path.join(os.path.dirname(__file__), 'repo')

    tmpdir: TemporaryDirectory
    repository: Repository

    def setUp(self):
        self.tmpdir = TemporaryDirectory(prefix='galileo_unittest_')
        self._prepare_repo()

        self.repository = Repository(self.tmpdir.name)

    def tearDown(self):
        self.tmpdir.cleanup()

    def _prepare_repo(self):
        for d in os.listdir(self.apps_dir):

            app_dir = os.path.join(self.apps_dir, d)
            with zipfile.ZipFile(os.path.join(self.tmpdir.name, '%s.zip' % d), 'w') as zipfd:
                for root, dirs, files in os.walk(app_dir):
                    for file in files:
                        zipfd.write(os.path.join(root, file), file)


class RepositoryTest(unittest.TestCase):
    repo = TemporaryRepositoryResource()

    @classmethod
    def setUpClass(cls) -> None:
        cls.repo.setUp()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.repo.tearDown()

    def test_list_archives(self):
        archives = self.repo.repository.list_archives()

        self.assertEqual(3, len(archives), 'expected three archives, got %s' % archives)

        names = [path.split('/')[-1] for path in archives]
        self.assertIn('testapp.zip', names)
        self.assertIn('noname.zip', names)
        self.assertIn('nomanifest.zip', names)

    def test_list_apps(self):
        apps = self.repo.repository.list_apps()
        self.assertEqual(1, len(apps), 'expected one app, got %s' % apps)

        info = apps[0]

        self.assertEqual('testapp', info.name)
        self.assertEqual('testapp', info.manifest['name'])

    def test_get_app(self):
        info = self.repo.repository.get_app('testapp')

        self.assertEqual('testapp', info.name)
        self.assertEqual('testapp', info.manifest['name'])

    def test_get_app_nonexisting_app_returns_none(self):
        info = self.repo.repository.get_app('nonexisting')
        self.assertIsNone(info)

    def test_get_app_invalid_app_returns_none(self):
        info = self.repo.repository.get_app('noname')
        self.assertIsNone(info)

    def test_get_app_without_manifestreturns_none(self):
        info = self.repo.repository.get_app('nomanifest')
        self.assertIsNone(info)

    def test_add_and_get(self):
        repo = self.repo.repository
        info = repo.add(os.path.join(this_dir, 'testapp2.zip'))

        self.assertEqual('testapp2', info.name)
        self.assertEqual({'name': 'testapp2'}, info.manifest)

        actual = repo.get_app('testapp2')
        self.assertEqual(info, actual)

    def test_add_and_remove(self):
        repo = self.repo.repository
        info = repo.add(os.path.join(this_dir, 'testapp2.zip'))

        self.assertTrue(os.path.isfile(info.archive_path))

        deleted = repo.delete_app('testapp2')
        self.assertTrue(deleted)

        self.assertFalse(os.path.isfile(info.archive_path))


if __name__ == '__main__':
    unittest.main()



import random


class InvalidDistributionException(Exception):
    pass


def constant(x):
    return x


distributions = {
    'constant': constant,
    'uniform': random.uniform,
    'triangular': random.triangular,
    'normalvariate': random.normalvariate,
    'lognormvariate': random.lognormvariate,
    'expovariate': random.expovariate,
    'vonmisesvariate': random.vonmisesvariate,
    'gammavariate': random.gammavariate,
    'gauss': random.gauss,
    'betavariate': random.betavariate,
    'paretovariate': random.paretovariate,
    'weibullvariate': random.weibullvariate
}


def create_sampler(distribution: str, args: tuple):
    """
    Creates a generator for the given distribution with the given arguments.

    :param distribution: the distribution, e.g., 'lognormvariate'
    :param args: the arguments, e.g., (0.5, 1)
    :return: a generator
    """
    if distribution not in distributions:
        raise InvalidDistributionException('unknown distribution ' + distribution)

    try:
        fn = distributions[distribution]
    except TypeError as e:
        raise InvalidDistributionException('invalid distribution parameters: ' + str(e))

    if args is None:
        args = []

    while True:
        yield fn(*args)



from galileo.shell.shell import sleep

name = 'init'

__all__ = [
    'sleep'
]


def _init(module=None):
    from galileo.worker.context import Context
    from galileo.shell.shell import init_module
    import atexit
    import pymq

    ctx = Context()
    rds = ctx.create_redis()
    init_module(rds, name=module)

    atexit.register(pymq.shutdown)


_init('__main__')



import argparse
import logging
import os
import signal

import pymq
import redis
from galileodb.factory import create_experiment_database_from_env
from galileodb.recorder import ExperimentTelemetryRecorder
from pymq.provider.redis import RedisConfig

from galileo.controller import ExperimentController
from galileo.experiment.experimentd import ExperimentDaemon
from galileo.experiment.service import SimpleExperimentService

logger = logging.getLogger(__name__)


def main():
    signal.signal(signal.SIGTERM, handle_sigterm)
    parser = argparse.ArgumentParser()
    parser.add_argument('--logging', required=False,
                        help='set log level (DEBUG|INFO|WARN|...) to activate logging',
                        default=os.getenv('galileo_logging_level'))
    args = parser.parse_args()

    if args.logging:
        logging.basicConfig(level=logging._nameToLevel[args.logging])

    rds = redis.Redis(
        host=os.getenv('galileo_redis_host', 'localhost'),
        port=int(os.getenv('galileo_redis_port', 6379)),
        decode_responses=True
    )
    pymq.init(RedisConfig(rds))

    exp_db = create_experiment_database_from_env()
    exp_db.open()

    def recorder_factory(exp_id: str):
        return ExperimentTelemetryRecorder(rds, exp_db, exp_id)

    try:
        exp_controller = ExperimentController(rds)
        exp_service = SimpleExperimentService(exp_db)

        daemon = ExperimentDaemon(rds, recorder_factory, exp_controller, exp_service)
        daemon.run()
    finally:
        exp_db.close()


def handle_sigterm(*args):
    raise KeyboardInterrupt()


if __name__ == '__main__':
    main()



import os
import re
import time
from datetime import timedelta
from inspect import signature
from typing import List
from uuid import uuid4


def read_file(f, mode='rb'):
    """
    Convenience method for reading a file into a byte buffer.
    :param f: path to the file
    :return: a byte array
    """
    with open(f, mode=mode) as reader:
        return reader.read()


def mkdirp(path):
    """
    Creates a directory at the given path with all necessary parent directories. If the path is already a directory,
    nothing happens. If the path is a file, then a FileExistsError will be raised.
    :param path: the desired path
    :return: None
    """
    if not os.path.exists(path):
        os.makedirs(path)

    if os.path.isfile(path):
        raise FileExistsError("%s is an existing file" % path)


def uuid():
    """
    Returns a random UUID as string.

    :return: a UUID
    """
    return str(uuid4())


def namedtuples_from_strings(cls, ls: List[str]):
    params = list(signature(cls).parameters.values())
    tuples = list()
    for v in ls:
        args = v.split(',')
        typed = [params[i].annotation(args[i]) for i in range(len(args))]
        trace = cls(*typed)
        tuples.append(trace)
    return tuples


def subdict(data: dict, keys: list):
    return {k: data[k] for k in keys if k in data}


_time_units = {
    's': 'seconds',
    'm': 'minutes',
    'h': 'hours'
}
_time_pattern = re.compile('([0-9]+)([smh]?)')


def to_seconds(time_str: str) -> int:
    groups = _time_pattern.findall(time_str)

    kwargs = dict()
    for value, unit in groups:
        if not unit:
            unit = 's'
        kwargs[_time_units[unit]] = int(value)

    return round(timedelta(**kwargs).total_seconds())


def poll(condition, timeout=None, interval=0.5):
    remaining = 0
    if timeout is not None:
        remaining = timeout

    while not condition():
        if timeout is not None:
            remaining -= interval

            if remaining <= 0:
                raise TimeoutError('gave up waiting after %s seconds' % timeout)

        time.sleep(interval)



import os
import unittest

from galileo.apps.loader import AppClientLoader, AppClientDirectoryLoader


class AppLoaderTest(unittest.TestCase):
    loader: AppClientLoader

    def setUp(self) -> None:
        repo_dir = os.path.join(os.path.dirname(__file__), 'repo')
        self.loader = AppClientDirectoryLoader(repo_dir)

    def test_list(self):
        apps = self.loader.list()
        self.assertEqual(1, len(apps))
        self.assertEqual('testapp', apps[0].name)

    def test_load_non_existing_app(self):
        self.assertRaises(ValueError, self.loader.load, 'nonexisting')

    def test_load_dir_without_manifest(self):
        self.assertRaises(ValueError, self.loader.load, 'nomanifest')

    def test_load_dir_without_app_name(self):
        self.assertRaises(ValueError, self.loader.load, 'noname')

    def test_load_and_run(self):
        app = self.loader.load('testapp')

        self.assertEqual('testapp', app.name)
        self.assertEqual('galileo.clients.testapp', app.module.__name__)

        request = app.next_request()
        self.assertEqual('testapp', request.app_name)
        self.assertEqual('get', request.method)
        self.assertEqual('/test', request.endpoint)
        self.assertEqual({'counter': 1}, request.kwargs)

        request = app.next_request()
        self.assertEqual('testapp', request.app_name)
        self.assertEqual('get', request.method)
        self.assertEqual('/test', request.endpoint)
        self.assertEqual({'counter': 2}, request.kwargs)


if __name__ == '__main__':
    unittest.main()



import unittest

from galileo.controller.cluster import pack


class PackTest(unittest.TestCase):
    def test_pack_single(self):
        workers = ['a', 'b', 'c']
        clients = [1, 2, 3]

        result = pack(1, workers, clients)

        expected = {
            'a': 1
        }

        self.assertEqual(expected, result)

    def test_pack_multiple(self):
        workers = ['a', 'b', 'c']
        clients = [5, 2, 1]

        result = pack(5, workers, clients)

        expected = {
            'b': 2,
            'c': 3
        }

        self.assertEqual(expected, result)

    def test_pack_tie(self):
        workers = ['a', 'b', 'c']
        clients = [2, 1, 1]

        result = pack(1, workers, clients)

        expected = {
            'b': 1,
        }

        self.assertEqual(expected, result)



from typing import NamedTuple


class AppInfo(NamedTuple):
    name: str
    manifest: dict
    archive_path: str = None


class Context:
    name: str
    path: str
    manifest: dict
    parameters: dict

    def __str__(self) -> str:
        return 'Context%s' % self.__dict__

    def __repr__(self):
        return self.__str__()



import unittest.mock

import falcon
import pymq
import redis
from falcon import testing
from galileodb.sql.adapter import ExperimentSQLDatabase
from pymq.provider.redis import RedisConfig

from galileo.controller import ExperimentController, RedisClusterController
from galileo.experiment.service import SimpleExperimentService
from galileo.webapp.app import setup, CORSComponent, AppContext
from tests.testutils import RedisResource, SqliteResource


class ResourceTest(testing.TestCase):
    redis_resource: RedisResource = RedisResource()
    ctx: AppContext

    def setUp(self):
        super(ResourceTest, self).setUp()
        rds = self.init_rds()
        db = self.init_db()
        self.ctx = self.init_context(db, rds)
        self.app = self.create_api(self.ctx)

    def tearDown(self) -> None:
        super(ResourceTest, self).tearDown()
        pymq.shutdown()
        self.redis_resource.tearDown()
        self.db_resource.tearDown()

    def init_db(self):
        self.db_resource = SqliteResource()
        self.db_resource.setUp()
        return self.db_resource.db

    def init_rds(self):
        self.redis_resource.setUp()
        pymq.init(RedisConfig(self.redis_resource.rds))
        return self.redis_resource.rds

    def create_api(self, ctx) -> falcon.API:
        api = falcon.API(middleware=[CORSComponent()])
        setup(api, ctx)
        return api

    @staticmethod
    def init_context(db: ExperimentSQLDatabase, rds: redis.Redis):
        context = AppContext()

        context.rds = rds

        context.cctrl = RedisClusterController(context.rds)
        context.ectrl = ExperimentController(context.rds)
        context.exp_db = db
        context.exp_service = SimpleExperimentService(context.exp_db)
        context.repository = unittest.mock.MagicMock('repository')

        return context



import logging
import unittest

import pymq
from galileodb.model import ExperimentConfiguration, WorkloadConfiguration, QueuedExperiment, Experiment
from pymq.provider.redis import RedisConfig

from galileo.controller import ClusterController, ExperimentController, RedisClusterController
from tests.testutils import RedisResource

logging.basicConfig(level=logging.DEBUG)


class TestExperimentDaemon(unittest.TestCase):
    redis_resource: RedisResource = RedisResource()
    ectl: ExperimentController
    cctl: ClusterController

    def setUp(self) -> None:
        self.rds = self.init_rds()
        pymq.init(RedisConfig(self.rds))
        self.ectl = ExperimentController(self.rds)
        self.cctl = RedisClusterController(self.rds)
        self.rds.delete(RedisClusterController.worker_key, ExperimentController.queue_key)

    def tearDown(self) -> None:
        self.rds.delete(RedisClusterController.worker_key, ExperimentController.queue_key)
        pymq.shutdown()
        self.redis_resource.tearDown()

    def test_queue(self):
        self.cctl.register_worker('host1')

        exp = Experiment(name='my-experiment', creator='unittest')
        config = ExperimentConfiguration(2, 1, [WorkloadConfiguration('aservice', [1, 2], 2, 'constant')])
        self.ectl.queue(config, exp)

        queue = pymq.queue(ExperimentController.queue_key)

        self.assertEqual(1, queue.qsize())
        self.assertEqual(1, self.ectl.experiment_queue.qsize())

        queued_experiment: QueuedExperiment = queue.get()
        self.assertEqual(config, queued_experiment.configuration)
        self.assertEqual('my-experiment', queued_experiment.experiment.name)
        self.assertEqual('unittest', queued_experiment.experiment.creator)

    def test_cancel(self):
        self.cctl.register_worker('host1')

        exp_id = 'abcd'
        exp = Experiment(id=exp_id, name='my-experiment', creator='unittest')
        config = ExperimentConfiguration(2, 1, [])
        self.ectl.queue(config, exp)

        exp = Experiment(id='abcdef', name='my-experiment-2', creator='unittest')
        config = ExperimentConfiguration(2, 1, [])
        self.ectl.queue(config, exp)

        self.assertEqual(2, self.ectl.experiment_queue.qsize())
        cancelled = self.ectl.cancel(exp_id)
        self.assertTrue(cancelled)
        self.assertEqual(1, self.ectl.experiment_queue.qsize())

        queued = self.ectl.experiment_queue.get()
        self.assertEqual('abcdef', queued.experiment.id, 'expected to find experiment that was not cancelled')

    def init_rds(self):
        self.redis_resource.setUp()
        return self.redis_resource.rds


if __name__ == '__main__':
    unittest.main()



import os
import time
import unittest
from queue import Queue
from typing import NamedTuple
from unittest.mock import patch

from pymq.provider.simple import SimpleEventBus

from galileo.routing import ServiceRequest, RedisRoutingTable, RoutingRecord
from galileo.worker.api import ClientDescription, ClientConfig
from galileo.worker.client import Client
from galileo.worker.context import Context
from tests.testutils import RedisResource
from tests.worker.test_client import StaticRequestGenerator


class TestSingleRequest(unittest.TestCase):
    n = 10000

    redis_resource = RedisResource()

    def setUp(self) -> None:
        self.redis_resource.setUp()
        self.rds = self.redis_resource.rds
        self.rtbl = RedisRoutingTable(self.rds)

    def tearDown(self) -> None:
        self.redis_resource.tearDown()

    def test_debug_router(self):
        env = dict(os.environ)
        env['galileo_router_type'] = 'DebugRouter'

        client_id = 'unittest_client'
        ctx = Context(env)
        trace_queue = Queue()

        description = ClientDescription(client_id, 'unittest_worker', ClientConfig('aservice'))
        # ctx: Context, trace_queue: Queue, description: ClientDescription

        client = Client(ctx, trace_queue, description, eventbus=SimpleEventBus())
        client.request_generator = StaticRequestGenerator([ServiceRequest('aservice') for _ in range(self.n)])

        then = time.time()
        client.run()
        now = time.time()

        total = now - then

        print('DebugRouter: %.2f req/sec (%.4fs total)' % (self.n / total, total))

    @patch('galileo.routing.router.requests.request')
    def test_symmetry_router(self, mock_request):
        # mock http server
        Response = NamedTuple('Response', status_code=int, text=str, url=str, method=str)

        def fake_response(method, url):
            return Response(200, 'ok', url, method)

        mock_request.side_effect = fake_response

        self.rtbl.set_routing(RoutingRecord('aservice', ['host1', 'host2', 'host3'], [1, 2, 3]))

        env = dict(os.environ)
        env['galileo_router_type'] = 'CachingSymmetryHostRouter'

        client_id = 'unittest_client'
        ctx = Context(env)
        ctx.create_redis = lambda: self.rds
        trace_queue = Queue()

        description = ClientDescription(client_id, 'unittest_worker', ClientConfig('aservice'))
        # ctx: Context, trace_queue: Queue, description: ClientDescription

        client = Client(ctx, trace_queue, description, eventbus=SimpleEventBus())
        client.request_generator = StaticRequestGenerator([ServiceRequest('aservice') for _ in range(self.n)])

        then = time.time()
        client.run()
        now = time.time()

        total = now - then

        print('CachingSymmetryHostRouter: %.2f req/sec (%.4f total)' % (self.n / total, total))



from galileo.routing.balancer import Balancer, WeightedRoundRobinBalancer, StaticLocalhostBalancer, \
    WeightedRandomBalancer, StaticHostBalancer
from galileo.routing.router import ServiceRequest, Router, StaticRouter, HostRouter, ServiceRouter
from galileo.routing.table import RoutingRecord, RoutingTable, RedisRoutingTable, ReadOnlyListeningRedisRoutingTable

__all__ = [
    'ServiceRequest',
    'Router',
    'StaticRouter',
    'HostRouter',
    'ServiceRouter',
    'RoutingRecord',
    'RoutingTable',
    'RedisRoutingTable',
    'ReadOnlyListeningRedisRoutingTable',
    'Balancer',
    'WeightedRoundRobinBalancer',
    'StaticLocalhostBalancer',
    'WeightedRandomBalancer',
    'StaticHostBalancer'
]
